"""Tests for events.py."""
import pytest

import functools
import os
import signal
import subprocess
import sys
import unittest

import asyncio
from trio_asyncio import aio_as_trio

from . import utils as test_utils

pytestmark = pytest.mark.skipif(sys.platform == 'win32',
                                reason="Not supported on Windows")

class MySubprocessProtocol(asyncio.SubprocessProtocol):
    def __init__(self, loop):
        self.state = 'INITIAL'
        self.transport = None
        self.connected = asyncio.Future(loop=loop)
        self.completed = asyncio.Future(loop=loop)
        self.disconnects = {fd: asyncio.Future(loop=loop) for fd in range(3)}
        self.data = {1: b'', 2: b''}
        self.returncode = None
        self.got_data = {1: asyncio.Event(loop=loop), 2: asyncio.Event(loop=loop)}

    def connection_made(self, transport):
        self.transport = transport
        assert self.state == 'INITIAL', self.state
        self.state = 'CONNECTED'
        self.connected.set_result(None)

    def connection_lost(self, exc):
        assert self.state == 'CONNECTED', self.state
        self.state = 'CLOSED'
        self.completed.set_result(None)

    def pipe_data_received(self, fd, data):
        assert self.state == 'CONNECTED', self.state
        self.data[fd] += data
        self.got_data[fd].set()

    def pipe_connection_lost(self, fd, exc):
        assert self.state == 'CONNECTED', self.state
        if exc:
            self.disconnects[fd].set_exception(exc)
        else:
            self.disconnects[fd].set_result(exc)

    def process_exited(self):
        assert self.state == 'CONNECTED', self.state
        self.returncode = self.transport.get_returncode()


def check_terminated(returncode):
    if sys.platform == 'win32':
        assert isinstance(returncode, int)
        # expect 1 but sometimes get 0
    else:
        assert -signal.SIGTERM == returncode


def check_killed(returncode):
    if sys.platform == 'win32':
        assert isinstance(returncode, int)
        # expect 1 but sometimes get 0
    else:
        assert -signal.SIGKILL == returncode


@pytest.mark.trio
async def test_subprocess_exec(loop):
    await run_subprocess_exec(loop)


@aio_as_trio
async def run_subprocess_exec(loop):
    prog = os.path.join(os.path.dirname(__file__), 'scripts', 'echo.py')

    connect = loop.subprocess_exec(
        functools.partial(MySubprocessProtocol, loop), sys.executable, prog
    )
    transp, proto = await connect
    assert isinstance(proto, MySubprocessProtocol)
    await proto.connected
    assert 'CONNECTED' == proto.state

    stdin = transp.get_pipe_transport(0)
    stdin.write(b'Python The Winner')
    await proto.got_data[1].wait()
    with test_utils.disable_logger():
        transp.close()
    await proto.completed
    check_killed(proto.returncode)
    assert b'Python The Winner' == proto.data[1]


@pytest.mark.trio
async def test_subprocess_interactive(loop):
    await run_subprocess_interactive(loop)


@aio_as_trio
async def run_subprocess_interactive(loop):
    prog = os.path.join(os.path.dirname(__file__), 'scripts', 'echo.py')

    connect = loop.subprocess_exec(
        functools.partial(MySubprocessProtocol, loop), sys.executable, prog
    )
    transp, proto = await connect
    assert isinstance(proto, MySubprocessProtocol)
    await proto.connected
    assert 'CONNECTED' == proto.state

    stdin = transp.get_pipe_transport(0)
    stdin.write(b'Python ')
    await proto.got_data[1].wait()
    proto.got_data[1].clear()
    assert b'Python ' == proto.data[1]

    stdin.write(b'The Winner')
    await proto.got_data[1].wait()
    assert b'Python The Winner' == proto.data[1]

    with test_utils.disable_logger():
        transp.close()
    await proto.completed
    check_killed(proto.returncode)


@pytest.mark.trio
async def test_subprocess_shell(loop):
    await run_subprocess_shell(loop)


@aio_as_trio
async def run_subprocess_shell(loop):
    connect = loop.subprocess_shell(functools.partial(MySubprocessProtocol, loop), 'echo Python')
    transp, proto = await connect
    assert isinstance(proto, MySubprocessProtocol)
    await proto.connected

    transp.get_pipe_transport(0).close()
    await proto.completed
    assert 0 == proto.returncode
    assert all(f.done() for f in proto.disconnects.values())
    assert proto.data[1].rstrip(b'\r\n') == b'Python'
    assert proto.data[2] == b''
    transp.close()


@pytest.mark.trio
async def test_subprocess_exitcode(loop):
    await run_subprocess_exitcode(loop)


@aio_as_trio
async def run_subprocess_exitcode(loop):
    connect = loop.subprocess_shell(
        functools.partial(MySubprocessProtocol, loop),
        'exit 7',
        stdin=None,
        stdout=None,
        stderr=None
    )
    transp, proto = await connect
    assert isinstance(proto, MySubprocessProtocol)
    await proto.completed
    assert 7 == proto.returncode
    transp.close()


@pytest.mark.trio
async def test_subprocess_close_after_finish(loop):
    await run_subprocess_close_after_finish(loop)


@aio_as_trio
async def run_subprocess_close_after_finish(loop):
    connect = loop.subprocess_shell(
        functools.partial(MySubprocessProtocol, loop),
        'exit 7',
        stdin=None,
        stdout=None,
        stderr=None
    )
    transp, proto = await connect
    assert isinstance(proto, MySubprocessProtocol)
    assert transp.get_pipe_transport(0) is None
    assert transp.get_pipe_transport(1) is None
    assert transp.get_pipe_transport(2) is None
    await proto.completed
    assert 7 == proto.returncode
    assert transp.close() is None


@pytest.mark.trio
async def test_subprocess_kill(loop):
    await run_subprocess_kill(loop)


@aio_as_trio
async def run_subprocess_kill(loop):
    prog = os.path.join(os.path.dirname(__file__), 'scripts', 'echo.py')

    connect = loop.subprocess_exec(
        functools.partial(MySubprocessProtocol, loop), sys.executable, prog
    )
    transp, proto = await connect
    assert isinstance(proto, MySubprocessProtocol)
    await proto.connected

    transp.kill()
    await proto.completed
    check_killed(proto.returncode)
    transp.close()


@pytest.mark.trio
async def test_subprocess_terminate(loop):
    await run_subprocess_terminate(loop)


@aio_as_trio
async def run_subprocess_terminate(loop):
    prog = os.path.join(os.path.dirname(__file__), 'scripts', 'echo.py')

    connect = loop.subprocess_exec(
        functools.partial(MySubprocessProtocol, loop), sys.executable, prog
    )
    transp, proto = await connect
    assert isinstance(proto, MySubprocessProtocol)
    await proto.connected

    transp.terminate()
    await proto.completed
    check_terminated(proto.returncode)
    transp.close()


@unittest.skipIf(sys.platform == 'win32', "Don't have SIGHUP")
@pytest.mark.trio
async def test_subprocess_send_signal(loop):
    await run_subprocess_send_signal(loop)


@aio_as_trio
async def run_subprocess_send_signal(loop):
    # bpo-31034: Make sure that we get the default signal handler (killing
    # the process). The parent process may have decided to ignore SIGHUP,
    # and signal handlers are inherited.
    old_handler = signal.signal(signal.SIGHUP, signal.SIG_DFL)
    try:
        prog = os.path.join(os.path.dirname(__file__), 'scripts', 'echo.py')

        connect = loop.subprocess_exec(
            functools.partial(MySubprocessProtocol, loop), sys.executable, prog
        )
        transp, proto = await connect
        assert isinstance(proto, MySubprocessProtocol)
        await proto.connected

        transp.send_signal(signal.SIGHUP)
        await proto.completed
        assert -signal.SIGHUP == proto.returncode
        transp.close()
    finally:
        signal.signal(signal.SIGHUP, old_handler)


@pytest.mark.trio
async def test_subprocess_stderr(loop):
    await run_subprocess_stderr(loop)


@aio_as_trio
async def run_subprocess_stderr(loop):
    prog = os.path.join(os.path.dirname(__file__), 'scripts', 'echo2.py')

    connect = loop.subprocess_exec(
        functools.partial(MySubprocessProtocol, loop), sys.executable, prog
    )
    transp, proto = await connect
    assert isinstance(proto, MySubprocessProtocol)
    await proto.connected

    stdin = transp.get_pipe_transport(0)
    stdin.write(b'test')

    await proto.completed

    transp.close()
    assert b'OUT:test' == proto.data[1]
    assert proto.data[2].startswith(b'ERR:test'), proto.data[2]
    assert 0 == proto.returncode


@pytest.mark.trio
async def test_subprocess_stderr_redirect_to_stdout(loop):
    await run_subprocess_stderr_redirect_to_stdout(loop)


@aio_as_trio
async def run_subprocess_stderr_redirect_to_stdout(loop):
    prog = os.path.join(os.path.dirname(__file__), 'scripts', 'echo2.py')

    connect = loop.subprocess_exec(
        functools.partial(MySubprocessProtocol, loop),
        sys.executable,
        prog,
        stderr=subprocess.STDOUT
    )
    transp, proto = await connect
    assert isinstance(proto, MySubprocessProtocol)
    await proto.connected

    stdin = transp.get_pipe_transport(0)
    assert transp.get_pipe_transport(1) is not None
    assert transp.get_pipe_transport(2) is None

    stdin.write(b'test')
    await proto.completed
    assert proto.data[1].startswith(b'OUT:testERR:test'), proto.data[1]
    assert b'' == proto.data[2]

    transp.close()
    assert 0 == proto.returncode


@pytest.mark.trio
async def test_subprocess_close_client_stream(loop):
    await run_subprocess_close_client_stream(loop)


@aio_as_trio
async def run_subprocess_close_client_stream(loop):
    prog = os.path.join(os.path.dirname(__file__), 'scripts', 'echo3.py')

    connect = loop.subprocess_exec(
        functools.partial(MySubprocessProtocol, loop), sys.executable, prog
    )
    transp, proto = await connect
    assert isinstance(proto, MySubprocessProtocol)
    await proto.connected

    stdin = transp.get_pipe_transport(0)
    stdout = transp.get_pipe_transport(1)
    stdin.write(b'test')
    await proto.got_data[1].wait()
    assert b'OUT:test' == proto.data[1]

    stdout.close()
    await proto.disconnects[1]
    stdin.write(b'xxx')
    await proto.got_data[2].wait()
    if sys.platform != 'win32':
        assert b'ERR:BrokenPipeError' == proto.data[2]
    else:
        # After closing the read-end of a pipe, writing to the
        # write-end using os.write() fails with errno==EINVAL and
        # GetLastError()==ERROR_INVALID_NAME on Windows!?!  (Using
        # WriteFile() we get ERROR_BROKEN_PIPE as expected.)
        assert b'ERR:OSError' == proto.data[2]
    with test_utils.disable_logger():
        transp.close()
    await proto.completed
    check_killed(proto.returncode)


@pytest.mark.trio
async def test_subprocess_wait_no_same_group(loop):
    await run_subprocess_wait_no_same_group(loop)


@aio_as_trio
async def run_subprocess_wait_no_same_group(loop):
    # start the new process in a new session
    connect = loop.subprocess_shell(
        functools.partial(MySubprocessProtocol, loop),
        'exit 7',
        stdin=None,
        stdout=None,
        stderr=None,
        start_new_session=True
    )
    transp, proto = await connect
    assert isinstance(proto, MySubprocessProtocol)
    await proto.completed
    assert 7 == proto.returncode
    with test_utils.disable_logger():
        transp.close()


@pytest.mark.trio
async def test_subprocess_exec_invalid_args(loop):
    @aio_as_trio
    async def connect(**kwds):
        await loop.subprocess_exec(asyncio.SubprocessProtocol, 'pwd', **kwds)

    with pytest.raises(ValueError):
        await connect(universal_newlines=True)
    with pytest.raises(ValueError):
        await connect(bufsize=4096)
    with pytest.raises(ValueError):
        await connect(shell=True)


@pytest.mark.trio
async def test_subprocess_shell_invalid_args(loop):
    @aio_as_trio
    async def connect(cmd=None, **kwds):
        if not cmd:
            cmd = 'pwd'
        await loop.subprocess_shell(asyncio.SubprocessProtocol, cmd, **kwds)

    with pytest.raises(ValueError):
        await connect(['ls', '-l'])
    with pytest.raises(ValueError):
        await connect(universal_newlines=True)
    with pytest.raises(ValueError):
        await connect(bufsize=4096)
    with pytest.raises(ValueError):
        await connect(shell=False)



import os
import sys
import warnings
import py
import pytest
import unittest

try:
    # the global 'test' package installed with Python
    from test import test_asyncio
except ImportError:
    warnings.warn(
        "Can't run the Python asyncio tests because they're not installed. "
        "On a Debian/Ubuntu system, you might need to install the "
        "libpython{}.{}-testsuite package.".format(*sys.version_info[:2]),
        RuntimeWarning,
    )
else:
    # threading_cleanup() causes a 5-second or so delay (100
    # gc.collect()'s) and warning on any test that holds a reference
    # to the event loop in the testsuite class, because
    # SyncTrioEventLoop spawns a thread that only exits when the
    # loop is closed. Nerf it.
    from test.support import threading_cleanup

    def threading_no_cleanup(*original_values):
        pass

    threading_cleanup.__code__ = threading_no_cleanup.__code__

    asyncio_test_dir = py.path.local(test_asyncio.__path__[0])

    def aio_test_nodeid(fspath):
        relpath = fspath.relto(asyncio_test_dir)
        if relpath:
            return "/Python-{}.{}/test_asyncio/".format(*sys.version_info[:2]) + relpath
        return None


    # A pytest.Module that will only collect unittest.TestCase
    # classes, so that we don't get spurious warnings about things
    # like TestSelector and TestEventLoop (which are fakes used by the
    # tests, directly containing no tests themselves) not being collectable.
    class UnittestOnlyModule(pytest.Module):
        def istestclass(self, obj, name):
            return isinstance(obj, unittest.TestCase)


    # A pytest.Package whose only purpose is to mark that its children should
    # become UnittestOnlyModules (or other UnittestOnlyPackages).
    class UnittestOnlyPackage(pytest.Package):
        def collect(self):
            for node in super().collect():
                if isinstance(node, pytest.Package):
                    node.__class__ = UnittestOnlyPackage
                elif isinstance(node, pytest.Module):
                    node.__class__ = UnittestOnlyModule
                node._nodeid = node._nodeid.replace("/__init__.py::", "/")
                yield node


    @pytest.hookimpl(tryfirst=True)
    def pytest_pycollect_makemodule(path, parent):
        # Convert collection of trio-asyncio tests.python into collection of
        # test.test_asyncio from the Python standard library.
        candidate = str(path.realpath())
        expected = os.path.realpath(
            os.path.join(os.path.dirname(__file__), "__init__.py")
        )
        if candidate == expected:
            fspath = py.path.local(test_asyncio.__file__)
            return UnittestOnlyPackage(fspath, parent, nodeid=aio_test_nodeid(fspath))


    def pytest_collection_modifyitems(items):
        by_id = {item.nodeid: item for item in items}
        aio_test_root = aio_test_nodeid(asyncio_test_dir / "foo")[:-3]

        def mark(marker, rel_id):
            try:
                by_id[aio_test_root + rel_id].add_marker(marker)
            except KeyError:
                warnings.warn(
                    "Tried to add marker {} to {}, but that test doesn't exist."
                    .format(marker, rel_id),
                    RuntimeWarning,
                    stacklevel=3,
                )

        def xfail(rel_id):
            mark(pytest.mark.xfail, rel_id)

        def skip(rel_id):
            mark(pytest.mark.skip, rel_id)

        # This hangs, probably due to the thread shenanigans (it works
        # fine with a greenlet-based sync loop)
        skip("test_base_events.py::RunningLoopTests::test_running_loop_within_a_loop")

        # Remainder of these have unclear issues
        if sys.version_info < (3, 8):
            xfail(
                "test_base_events.py::BaseEventLoopWithSelectorTests::"
                "test_log_slow_callbacks"
            )

        if sys.version_info >= (3, 7):
            xfail(
                "test_tasks.py::RunCoroutineThreadsafeTests::"
                "test_run_coroutine_threadsafe_task_factory_exception"
            )
        if sys.version_info >= (3, 8):
            xfail(
                "test_tasks.py::RunCoroutineThreadsafeTests::"
                "test_run_coroutine_threadsafe_task_cancelled"
            )
            xfail(
                "test_tasks.py::RunCoroutineThreadsafeTests::"
                "test_run_coroutine_threadsafe_with_timeout"
            )

        # These fail with ConnectionResetError on Pythons <= 3.7.x
        # for some unknown x. (3.7.1 fails, 3.7.5 and 3.7.6 pass;
        # older 3.6.x also affected, and older-or-all 3.5.x)
        if sys.platform != "win32" and sys.version_info < (3, 8):
            import selectors

            kinds = ("Select",)
            for candidate in ("Kqueue", "Epoll", "Poll"):
                if hasattr(selectors, candidate + "Selector"):
                    kinds += (candidate.replace("Epoll", "EPoll"),)
            for kind in kinds:
                tests = (
                    "test_create_ssl_connection", "test_create_ssl_unix_connection"
                )
                if sys.version_info < (3, 7):
                    tests += (
                        "test_legacy_create_ssl_connection",
                        "test_legacy_create_ssl_unix_connection",
                    )
                    stream_suite = "StreamReaderTests"
                else:
                    stream_suite = "StreamTests"
                for test in tests:
                    xfail("test_events.py::{}EventLoopTests::{}".format(kind, test))
                for which in ("open_connection", "open_unix_connection"):
                    xfail(
                        "test_streams.py::{}::test_{}_no_loop_ssl"
                        .format(stream_suite, which)
                    )



import pytest
import trio_asyncio
import asyncio
import trio
import sys
from . import utils as test_utils

# from tests import aiotest


class Seen:
    flag = 0


class TestMisc:
    @pytest.mark.trio
    async def test_close_no_stop(self):
        with pytest.raises(RuntimeError):
            async with trio_asyncio.open_loop() as loop:

                def close_no_stop():
                    loop.close()

                loop.call_soon(close_no_stop)

                await trio.sleep(0.1)
                await loop.wait_closed()

    @pytest.mark.trio
    async def test_too_many_stops(self):
        with trio.move_on_after(1) as scope:
            async with trio_asyncio.open_loop() as loop:
                await trio.hazmat.checkpoint()
                loop.stop()
        assert not scope.cancelled_caught, \
            "Possible deadlock after manual call to loop.stop"

    @pytest.mark.trio
    async def test_err1(self, loop):
        async def raise_err():
            raise RuntimeError("Foo")

        with pytest.raises(RuntimeError) as err:
            await trio_asyncio.aio_as_trio(raise_err, loop=loop)()
        assert err.value.args[0] == "Foo"

    @pytest.mark.trio
    async def test_err1_depr(self, loop):
        async def raise_err():
            raise RuntimeError("Foo")

        with test_utils.deprecate(self):
            with pytest.raises(RuntimeError) as err:
                await loop.run_asyncio(raise_err)
        assert err.value.args[0] == "Foo"

    @pytest.mark.trio
    async def test_err3(self, loop):
        owch = 0

        async def nest():
            nonlocal owch
            owch = 1
            raise RuntimeError("Hello")

        async def call_nested():
            with pytest.raises(RuntimeError) as err:
                await trio_asyncio.trio_as_aio(nest, loop=loop)()
            assert err.value.args[0] == "Hello"

        await trio_asyncio.aio_as_trio(call_nested, loop=loop)()
        assert owch

    @pytest.mark.trio
    async def test_err3_depr(self, loop):
        owch = 0

        async def nest():
            nonlocal owch
            owch = 1
            raise RuntimeError("Hello")

        async def call_nested():
            with pytest.raises(RuntimeError) as err:
                await loop.run_trio(nest)
            assert err.value.args[0] == "Hello"

        with test_utils.deprecate(self):
            await loop.run_asyncio(call_nested)
        assert owch

    @pytest.mark.trio
    async def test_run(self, loop):
        owch = 0

        async def nest():
            await trio.sleep(0.01)
            nonlocal owch
            owch = 1

        async def call_nested():
            await trio_asyncio.trio_as_aio(nest, loop=loop)()

        await trio_asyncio.aio_as_trio(call_nested, loop=loop)()
        assert owch

    @pytest.mark.trio
    async def test_run_depr(self, loop):
        owch = 0

        async def nest():
            await trio.sleep(0.01)
            nonlocal owch
            owch = 1

        async def call_nested():
            with test_utils.deprecate(self):
                await loop.run_trio(nest)

        with test_utils.deprecate(self):
            await loop.run_asyncio(call_nested)
        assert owch

    async def _test_run(self):
        owch = 0

        async def nest():
            await trio.sleep(0.01)
            nonlocal owch
            owch = 1

        async def call_nested():
            await trio_asyncio.trio_as_aio(nest)()

        await trio_asyncio.aio_as_trio(call_nested)()
        assert owch

    def test_run2(self):
        trio_asyncio.run(self._test_run)

    async def _test_run_depr(self):
        owch = 0

        async def nest():
            await trio.sleep(0.01)
            nonlocal owch
            owch = 1

        async def call_nested():
            await trio_asyncio.run_trio(nest)

        with test_utils.deprecate(self):
            await trio_asyncio.run_asyncio(call_nested)
        assert owch

    def test_run2_depr(self):
        trio_asyncio.run(self._test_run_depr)

    @pytest.mark.trio
    async def test_run_task(self):
        owch = 0

        async def nest(x):
            nonlocal owch
            owch += x

        with pytest.raises(RuntimeError):
            trio_asyncio.run_trio_task(nest, 100)

        with pytest.raises(RuntimeError):
            with trio_asyncio.open_loop():
                nest(1000)

        async with trio_asyncio.open_loop():
            trio_asyncio.run_trio_task(nest, 1)
        await trio.sleep(0.05)
        assert owch == 1

    @pytest.mark.trio
    async def test_err2(self, loop):
        owch = 0

        async def nest():
            nonlocal owch
            owch = 1
            raise RuntimeError("Hello")

        async def call_nested():
            await trio_asyncio.aio_as_trio(nest, loop=loop)()

        async def call_more_nested():
            with pytest.raises(RuntimeError) as err:
                await trio_asyncio.trio_as_aio(call_nested, loop=loop)()
            assert err.value.args[0] == "Hello"

        await trio_asyncio.aio_as_trio(call_more_nested, loop=loop)()
        assert owch

    @pytest.mark.trio
    async def test_err2_depr(self, loop):
        owch = 0

        async def nest():
            nonlocal owch
            owch = 1
            raise RuntimeError("Hello")

        async def call_nested():
            await loop.run_asyncio(nest)

        async def call_more_nested():
            with pytest.raises(RuntimeError) as err:
                await loop.run_trio(call_nested)
            assert err.value.args[0] == "Hello"

        with test_utils.deprecate(self):
            await loop.run_asyncio(call_more_nested)
        assert owch

    @pytest.mark.trio
    async def test_run3(self, loop):
        owch = 0

        async def nest():
            nonlocal owch
            owch = 1

        async def call_nested():
            await trio_asyncio.aio_as_trio(nest, loop=loop)()

        async def call_more_nested():
            await trio_asyncio.trio_as_aio(call_nested, loop=loop)()

        await trio_asyncio.aio_as_trio(call_more_nested, loop=loop)()
        assert owch

    @pytest.mark.trio
    async def test_run3_depr(self, loop):
        owch = 0

        async def nest():
            nonlocal owch
            owch = 1

        async def call_nested():
            await loop.run_asyncio(nest)

        async def call_more_nested():
            await loop.run_trio(call_nested)

        with test_utils.deprecate(self):
            await loop.run_asyncio(call_more_nested)
        assert owch

    @pytest.mark.trio
    async def test_cancel_sleep(self, loop):
        owch = 0

        def do_not_run():
            nonlocal owch
            owch = 1
            raise Exception("should not run")

        async def cancel_sleep():
            h = loop.call_later(0.2, do_not_run)
            await asyncio.sleep(0.1, loop=loop)
            h.cancel()
            await asyncio.sleep(0.3, loop=loop)

        await trio_asyncio.aio_as_trio(cancel_sleep, loop=loop)()
        assert owch == 0

    @pytest.mark.trio
    async def test_cancel_sleep_depr(self, loop):
        owch = 0

        def do_not_run():
            nonlocal owch
            owch = 1
            raise Exception("should not run")

        async def cancel_sleep():
            h = loop.call_later(0.2, do_not_run)
            await asyncio.sleep(0.1, loop=loop)
            h.cancel()
            await asyncio.sleep(0.3, loop=loop)

        with test_utils.deprecate(self):
            await loop.run_asyncio(cancel_sleep)
        assert owch == 0


@pytest.mark.trio
async def test_wrong_context_manager_order():
    take_down = trio.Event()
    async def work_in_asyncio():
        await asyncio.sleep(0)

    async def runner(*, task_status=trio.TASK_STATUS_IGNORED):
        await trio_asyncio.aio_as_trio(work_in_asyncio)()
        try:
            task_status.started()
            await take_down.wait()
        finally:
            await trio_asyncio.aio_as_trio(work_in_asyncio)()

    async with trio.open_nursery() as nursery:
        async with trio_asyncio.open_loop():
            await nursery.start(runner)
            take_down.set()


@pytest.mark.trio
@pytest.mark.skipif(sys.platform == 'win32',
                    reason="Not supported on Windows")
async def test_keyboard_interrupt_teardown():
    asyncio_loop_closed = trio.Event()

    async def work_in_trio_no_matter_what(*, task_status=trio.TASK_STATUS_IGNORED):
        await trio_asyncio.aio_as_trio(work_in_asyncio)()
        try:
            # KeyboardInterrupt won't cancel this coroutine thanks to the shield
            with trio.CancelScope(shield=True):
                task_status.started()
                await asyncio_loop_closed.wait()
        finally:
            # Hence this call will be exceuted after run_asyncio_loop is cancelled
            with pytest.raises(RuntimeError):
                await trio_asyncio.aio_as_trio(work_in_asyncio)()

    async def work_in_asyncio():
        await asyncio.sleep(0)

    async def run_asyncio_loop(nursery, *, task_status=trio.TASK_STATUS_IGNORED):
        with trio.CancelScope() as cancel_scope:
            try:
                async with trio_asyncio.open_loop():
                    # Starting a coroutine from here make it inherit the access
                    # to the asyncio loop context manager
                    await nursery.start(work_in_trio_no_matter_what)
                    task_status.started(cancel_scope)
                    await trio.sleep_forever()
            finally:
                asyncio_loop_closed.set()

    import signal
    import threading
    with pytest.raises(KeyboardInterrupt):
        with pytest.warns(RuntimeWarning):
            async with trio.open_nursery() as nursery:
                await nursery.start(run_asyncio_loop, nursery)
                # Trigger KeyboardInterrupt that should propagate accross the coroutines
                signal.pthread_kill(threading.get_ident(), signal.SIGINT)



"""Utilities shared by tests."""

import sys
import pytest
import contextlib
import logging
from asyncio.log import logger

from trio_asyncio import TrioAsyncioDeprecationWarning

def deprecate(tc):
    return pytest.warns(TrioAsyncioDeprecationWarning)

def deprecate_stdlib(tc, vers=None):
    if vers is None or sys.version_info >= vers:
        return pytest.deprecated_call()

    class _deprecate:
        def __init__(self, tc):
            pass

        def __enter__(self):
            return self

        def __exit__(self, *tb):
            pass

    return _deprecate(tc)


@contextlib.contextmanager
def disable_logger():
    """Context manager to disable asyncio logger.

    For example, it can be used to ignore warnings in debug mode.
    """
    old_level = logger.level
    try:
        logger.setLevel(logging.CRITICAL + 1)
        yield
    finally:
        logger.setLevel(old_level)



import pytest
from async_generator import async_generator, yield_
from trio_asyncio import aio_as_trio, trio_as_aio, allow_asyncio
from trio_asyncio import aio2trio, trio2aio
import asyncio
import trio
import sniffio
from tests import aiotest
import sys
import warnings
from async_generator import asynccontextmanager
from .. import utils as test_utils
from trio_asyncio import TrioAsyncioDeprecationWarning


def de_deprecate_converter(func):
    def wrapper(proc):
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', TrioAsyncioDeprecationWarning)
            return func(proc)

    return wrapper


class SomeThing:
    flag = 0

    def __init__(self, loop):
        self.loop = loop

    async def dly_trio(self):
        if sys.version_info >= (3, 7):
            assert sniffio.current_async_library() == "trio"
        await trio.sleep(0.01)
        self.flag |= 2
        return 8

    @trio_as_aio
    async def dly_trio_adapted(self):
        if sys.version_info >= (3, 7):
            assert sniffio.current_async_library() == "trio"
        await trio.sleep(0.01)
        self.flag |= 2
        return 8

    @de_deprecate_converter(aio2trio)
    async def dly_trio_depr(self):
        if sys.version_info >= (3, 7):
            assert sniffio.current_async_library() == "trio"
        await trio.sleep(0.01)
        self.flag |= 2
        return 8

    @de_deprecate_converter(trio2aio)
    async def dly_asyncio_depr(self):
        if sys.version_info >= (3, 7):
            assert sniffio.current_async_library() == "asyncio"
        await asyncio.sleep(0.01, loop=self.loop)
        self.flag |= 1
        return 4

    @aio_as_trio
    async def dly_asyncio_adapted(self):
        if sys.version_info >= (3, 7):
            assert sniffio.current_async_library() == "asyncio"
        await asyncio.sleep(0.01, loop=self.loop)
        self.flag |= 1
        return 4

    async def dly_asyncio(self, do_test=True):
        if do_test and sys.version_info >= (3, 7):
            assert sniffio.current_async_library() == "asyncio"
        await asyncio.sleep(0.01, loop=self.loop)
        self.flag |= 1
        return 4

    @async_generator
    async def iter_asyncio(self, do_test=True):
        if do_test and sys.version_info >= (3, 7):
            assert sniffio.current_async_library() == "asyncio"
        await asyncio.sleep(0.01, loop=self.loop)
        await yield_(1)
        await asyncio.sleep(0.01, loop=self.loop)
        await yield_(2)
        await asyncio.sleep(0.01, loop=self.loop)
        self.flag |= 1

    @async_generator
    async def iter_trio(self):
        if sys.version_info >= (3, 7):
            assert sniffio.current_async_library() == "trio"
        await trio.sleep(0.01)
        await yield_(1)
        await trio.sleep(0.01)
        await yield_(2)
        await trio.sleep(0.01)
        self.flag |= 1

    @asynccontextmanager
    @async_generator
    async def ctx_asyncio(self):
        await asyncio.sleep(0.01, loop=self.loop)
        self.flag |= 1
        await yield_(self)
        await asyncio.sleep(0.01, loop=self.loop)
        self.flag |= 2

    @asynccontextmanager
    @async_generator
    async def ctx_trio(self):
        await trio.sleep(0.01)
        self.flag |= 1
        await yield_(self)
        await trio.sleep(0.01)
        self.flag |= 2


class TestAdapt(aiotest.TestCase):
    @pytest.mark.trio
    async def test_asyncio_trio_depr(self, loop):
        """Call asyncio from trio"""

        sth = SomeThing(loop)
        res = await aio_as_trio(sth.dly_trio_depr, loop=loop)()
        assert res == 8
        assert sth.flag == 2

    @pytest.mark.trio
    async def test_asyncio_trio_adapted(self, loop):
        """Call asyncio from trio"""

        sth = SomeThing(loop)
        res = await aio_as_trio(sth.dly_trio_adapted, loop=loop)()
        assert res == 8
        assert sth.flag == 2

    @pytest.mark.trio
    async def test_asyncio_trio_depr2(self, loop):
        """Call asyncio from trio"""

        sth = SomeThing(loop)
        with test_utils.deprecate(self):
            res = await loop.run_asyncio(sth.dly_trio_adapted)
        assert res == 8
        assert sth.flag == 2

    @pytest.mark.trio
    async def test_trio_asyncio_adapted(self, loop):
        sth = SomeThing(loop)
        res = await sth.dly_asyncio_adapted()
        assert res == 4
        assert sth.flag == 1

    @pytest.mark.trio
    async def test_trio_asyncio(self, loop):
        sth = SomeThing(loop)
        res = await aio_as_trio(sth.dly_asyncio)()
        assert res == 4
        assert sth.flag == 1

    @pytest.mark.trio
    async def test_trio_asyncio_awaitable(self, loop):
        sth = SomeThing(loop)
        res = await aio_as_trio(sth.dly_asyncio())
        assert res == 4
        assert sth.flag == 1

    @pytest.mark.trio
    async def test_trio_asyncio_future(self, loop):
        sth = SomeThing(loop)
        f = sth.dly_asyncio(do_test=False)
        f = asyncio.ensure_future(f)
        res = await aio_as_trio(f)
        assert res == 4
        assert sth.flag == 1

    @pytest.mark.trio
    async def test_trio_asyncio_depr(self, loop):
        sth = SomeThing(loop)
        res = await sth.dly_asyncio_depr()
        assert res == 4
        assert sth.flag == 1

    @pytest.mark.trio
    async def test_trio_asyncio_iter(self, loop):
        sth = SomeThing(loop)
        n = 0
        if sys.version_info >= (3, 7):
            assert sniffio.current_async_library() == "trio"
        async for x in aio_as_trio(sth.iter_asyncio()):
            n += 1
            assert x == n
        assert n == 2
        assert sth.flag == 1

    async def run_asyncio_trio_iter(self, loop):
        sth = SomeThing(loop)
        n = 0
        if sys.version_info >= (3, 7):
            assert sniffio.current_async_library() == "asyncio"
        async for x in trio_as_aio(sth.iter_trio()):
            n += 1
            assert x == n
        assert n == 2
        assert sth.flag == 1

    @pytest.mark.trio
    async def test_asyncio_trio_iter(self, loop):
        await aio_as_trio(self.run_asyncio_trio_iter)(loop)

    @pytest.mark.trio
    async def test_trio_asyncio_ctx(self, loop):
        sth = SomeThing(loop)
        async with aio_as_trio(sth.ctx_asyncio()):
            assert sth.flag == 1
        assert sth.flag == 3

    async def run_asyncio_trio_ctx(self, loop):
        sth = SomeThing(loop)
        async with trio_as_aio(sth.ctx_trio()):
            assert sth.flag == 1
        assert sth.flag == 3

    @pytest.mark.trio
    async def test_asyncio_trio_ctx(self, loop):
        await aio_as_trio(self.run_asyncio_trio_ctx)(loop)


class TestAllow(aiotest.TestCase):
    async def run_asyncio_trio(self, loop):
        """Call asyncio from trio"""
        sth = SomeThing(loop)
        res = await trio_as_aio(sth.dly_trio, loop=loop)()
        assert res == 8
        assert sth.flag == 2

    async def run_asyncio_trio_adapted(self, loop):
        """Call asyncio from trio"""
        sth = SomeThing(loop)
        res = await sth.dly_trio_adapted()
        assert res == 8
        assert sth.flag == 2

    async def run_asyncio_trio_depr(self, loop):
        """Call asyncio from trio"""
        sth = SomeThing(loop)
        res = await sth.dly_trio_depr()
        assert res == 8
        assert sth.flag == 2

    @pytest.mark.trio
    async def test_asyncio_trio(self, loop):
        await allow_asyncio(self.run_asyncio_trio, loop)

    async def run_trio_asyncio(self, loop):
        sth = SomeThing(loop)
        res = await sth.dly_asyncio(do_test=False)
        assert res == 4
        assert sth.flag == 1

    @pytest.mark.trio
    async def test_trio_asyncio(self, loop):
        await allow_asyncio(self.run_trio_asyncio, loop)

    async def run_trio_asyncio_future(self, loop):
        sth = SomeThing(loop)
        f = sth.dly_asyncio(do_test=False)
        f = asyncio.ensure_future(f)
        res = await f
        assert res == 4
        assert sth.flag == 1

    @pytest.mark.trio
    async def test_trio_asyncio_future(self, loop):
        await allow_asyncio(self.run_trio_asyncio_future, loop)

    async def run_trio_asyncio_depr(self, loop):
        sth = SomeThing(loop)
        res = await sth.dly_asyncio_depr()
        assert res == 4
        assert sth.flag == 1

    @pytest.mark.trio
    async def test_trio_asyncio_depr(self, loop):
        await allow_asyncio(self.run_trio_asyncio_depr, loop)

    async def run_trio_asyncio_iter(self, loop):
        sth = SomeThing(loop)
        n = 0
        async for x in sth.iter_asyncio(do_test=False):
            n += 1
            assert x == n
        assert n == 2
        assert sth.flag == 1

    @pytest.mark.trio
    async def test_trio_asyncio_iter(self, loop):
        await allow_asyncio(self.run_trio_asyncio_iter, loop)

    async def run_trio_asyncio_ctx(self, loop):
        sth = SomeThing(loop)
        async with sth.ctx_asyncio():
            assert sth.flag == 1
        assert sth.flag == 3

    @pytest.mark.trio
    async def test_trio_asyncio_ctx(self, loop):
        await allow_asyncio(self.run_trio_asyncio_ctx, loop)
